#!/usr/bin/perl

# Author: Larry Doolittle <ldoolitt@recycle.lbl.gov>
# Initial construction: April, 2005
# Status as of 2005-04-15:
#    Works OK, needs more error handling, can duplicate USRP's gpif_usrp1.c if you
#    ignore flowstate.  Need suggestions for handling FIFOFlag, TCXpire, IntReady,
#    and non-CMOS CTLs.  Tested on debian sarge/sid, perl v5.8.4.

# Suggested Makefile usage:
#    foo.c: foo.gpif
#           gpif.pl $< > $@ || rm $@

sub error_msg {
	my ($msg) = @_;
	print STDERR "line $source_line: $msg\n";
	$err_count++;
}

sub parse_act {
	my ($a) = @_;
	my $result = $ctl_def{$a};
	error_msg("unknown action ($a)\n") if ($result eq "");
	$result = 1<<$result;
	# print "// parse_act input $a result $result\n";
	$result;
}

sub parse_term {
	my ($t) = @_;
	my $result = $rdy_def{$t};
	error_msg("unknown conditional term ($t)") if ($result eq "");
	# print "// parse_term($t) result($result)\n";
	$result;
}

sub parse_targ {
	my ($t) = @_;
	my $result = $t;
	$result = 7 if ($t eq "idle");
	# print "// parse_targ input $t result $result\n";
	$result;
}

# three operators, two arguments optionally negated, twelve combinations in all
# normal
$logic_lookup{"&"}=0;
$logic_lookup{"|"}=1;
$logic_lookup{"^"}=2;
$logic_lookup{"~^~"}=2;
$logic_lookup{"~&"}=3;
# polarity reversed
$logic_lookup{"~|~"}=4;
$logic_lookup{"~&~"}=5;
$logic_lookup{"~^"}=6;
$logic_lookup{"^~"}=6;
$logic_lookup{"|~"}=7;
# terms reversed
$logic_lookup{"&~"}=11;
# terms and polarity reversed
$logic_lookup{"~|"}=15;

sub parse_cond {
	my ($c) = @_;
	my ($op, $terma, $termb, $reverse);
	# print "// conditional ($c)\n";
	if ($c eq "0") {
		$terma=0;
		$op="3";   # ~A&B
		$termb=0;
		$reverse=0;
	} elsif ($c eq "1") {
		$terma=0;
		$op="3";   # ~A&B
		$termb=0;
		$reverse=1;
	} elsif ($c=~/^(~)?([\w#]+)$/) {
		# print "// singleton complement($1) $2\n";
		$terma=parse_term($2);
		$op="0";  # and
		$termb=$terma;
		$reverse=($1 eq "~");
	} elsif ($c=~/^(~)?\s*([\w#]+)\s*([&|^])\s*(~)?\s*([\w#]+)$/) {
		$ll=$logic_lookup{"$1$3$4"}+0;
		# print "// logic lookup($1$3$4)=$ll\n";
		error_msg("unavailable logic combination ($c)") if ($ll eq "");
		$terma=parse_term($2);
		$op=$ll&0x3;
		# not covered: 3 for ~A&B
		# could presuably cover this by allowing ~ in front of
		# either A or B, and rearranging to fit
		$termb=parse_term($5);
		$reverse=($ll&0x4)/4;
		if ($ll&0x8) {
			$foo=$termb;
			$termb=$terma;
			$terma=$foo;
		}
	} else {
		error_msg("unparsed conditional ($c)");
	}
	# print "// conditional($c) terma($terma) op($op) termb($termb) reverse($reverse)\n";
	($reverse, $op*64+$terma*8+$termb);
	# lfunc terma termb
}

sub parse_dest {
	my ($t,$e,$r) = @_;
	my $result=parse_targ($t)*8+parse_targ($e);  # high bit is reexecute
	$result+=128 if ($r=~/reex/);
	# print "// parse_dest result $result\n";
	$result;
}

sub memo_state {
	my ($w, $s, $act, $mot) = @_;
	my $reverse, $d0, $d1;
	# print "// found waveform $w state $s action $act motion $mot\n";
	for $a (split(/\s+/,$act)) {
		# print "//    action ($a)\n";
		   if ($a eq "activate") { $WaveData{$w.$s.1}|=2; }
		elsif ($a eq "next"    ) { $WaveData{$w.$s.1}|=4; }
		elsif ($a eq "incad"   ) { $WaveData{$w.$s.1}|=8; }
		elsif ($a eq "gint"    ) { $WaveData{$w.$s.1}|=16; }
		elsif ($a eq "sgl"     ) { $WaveData{$w.$s.1}|=32; }
		else  {$mask{$w.$s} |= parse_act($a);}
	}
	if ($mot=~/^wait\s*(\d+)\s*$/) {
		# print "// ======= wait($1)\n";
		$WaveData{$w.$s.0}=$1;
		# print "// WaveData  $w $s cond $WaveData{$w.$s.0}  dest $WaveData{$w.$s.2}\n";
	} elsif ($mot=~/^if\s*\(([^)]+)\)\s*then\s+([\w#]+)\s+else\s+([\w#]+)\s*(reex)?$/) {
		# print "// ======= if ($1) then ($2) else ($3)\n";
		$WaveData{$w.$s.1}|=1;  # DP=1
		($reverse,$WaveData{$w.$s.3})=parse_cond($1);
		if ($reverse) { $d0=$3; $d1=$2; }
		         else { $d0=$2; $d1=$3; }
		$WaveData{$w.$s.0}=parse_dest($d0,$d1,$4); # reex X branchon1 branchon0
		# print "// WaveData  $w $s cond $WaveData{$w.$s.0}  dest $WaveData{$w.$s.2}\n";
	} else {
		error_msg("unrecognized motion ($mot)");
	}
}

# Default for all WaveData are zero except LenBr
for ($w=0;$w<4;$w++) {
	for ($s=0; $s<7; $s++) {
		$WaveData{$w.$s.0}=1;
	}
}

$w=-1;  # invalid
while (<>) {
	$source_line++;
	chomp();
	$inline=$_;        # save for diagnostics
	s,//.*,,;          # strip // comments
	next if (!/\S/);   # ignore lines of whitespace only
	# print "// $_\n";
	if (/waveform\s*(\d*):\s*(.*)/i) {
		# print "found waveform $1 named $2\n";
		error_msg("invalid waveform number ($1)") if ($1<0 || $1>3);
		$w=$1;
	} elsif (/state\s*(\d*)\s*:\s*([^;]*);\s*(.*)/i) {
		error_msg("invalid state number ($1)") if ($1<0 || $1>6);
		memo_state($w, $1, $2, $3);
	} elsif (/^\s*RDY\s*([\d])+\s*=\s*([\w#]+)\s*$/i) {
		error_msg("invalid ready bit ($1)") if ($1<0 || $1>5);
		# print "// found ready($1) = $2\n";
		$rdy_def{$2}=$1;
	} elsif (/^\s*CTL\s*([\d])+\s*=\s*([\w#]+)(.*)/i) {
		error_msg("invalid control bit ($1)") if ($1<0 || $1>5);
		# print "// found control($1) = $2, rest = $3\n";
		$ctl_def{$2}=$1;
		$ctl_opt{$2}=$3;
		$bitn=$1;
		if ($3=~/\b1\b/ && $1>=0 && $1<=5) {
			$default_mask |= 1<<$bitn;
			# print "// default_mask now $default_mask\n";
		}
	} else {
		error_msg("unrecognized construct ($_)");
	}
}

@titles=("LenBr ", "Opcode", "Output", "LFun  ");

if ($err_count > 0) {
	print STDERR "$err_count errors, aborting\n";
	exit 1;
}

print "const char WaveData[128] =\n{\n";
for ($w=0;$w<4;$w++) {
	print "// Wave $w\n";
	# print "/* mask     ";
	for ($s=0; $s<7; $s++) {
		# printf "0x%.2x,   ", $mask{$w.$s};
		$WaveData{$w.$s.2}=$default_mask^$mask{$w.$s};
	}
	# print " */\n";
	for ($z=0; $z<4; $z++) {
		print "/* $titles[$z] */  ";
		for ($s=0; $s<7; $s++) {
			printf "0x%.2X,   ", $WaveData{$w.$s.$z};
		}
		print "0x00,\n";
	}
}
print "};\n";
